# generated by datamodel-codegen:
#   filename:  solution-analyzer.schema.json
#   timestamp: 2024-06-26T17:45:07+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueDegreeCelsius,
    TQuantityValueGramPerLiter,
    TQuantityValueMicrometer,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMilliliterPerLiter,
    TQuantityValueMillimeterOfMercury,
    TQuantityValueMillimolePerLiter,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValueMilliOsmolesPerKilogram,
    TQuantityValueNanometer,
    TQuantityValuePercent,
    TQuantityValuePH,
    TQuantityValuePicogramPerMilliliter,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TClass,
    TDatacube,
    TDateTimeStampValue,
    TStatisticDatumRole,
    TStringValue,
    TUnit,
)


@dataclass(kw_only=True)
class DiagnosticTraceDocumentItem:
    description: Any


@dataclass(kw_only=True)
class DiagnosticTraceAggregateDocument:
    diagnostic_trace_document: list[DiagnosticTraceDocumentItem] | None = None


@dataclass(kw_only=True)
class AdmCoreREC202403ManifestSchema:
    vocabulary: list[str]
    json_schemas: list[str]
    field_id: str | None = None
    field_type: str | None = None
    shapes: list[str] | None = None


@dataclass(kw_only=True)
class OrderedItem:
    field_index: int | None = None


@dataclass(kw_only=True)
class DataSourceDocumentItem:
    data_source_identifier: TStringValue
    data_source_feature: TStringValue
    field_index: int | None = None


@dataclass(kw_only=True)
class DataSourceAggregateDocument:
    data_source_document: list[DataSourceDocumentItem]


@dataclass(kw_only=True)
class ErrorDocumentItem:
    error: TStringValue
    error_feature: TStringValue | None = None


@dataclass(kw_only=True)
class ErrorAggregateDocument:
    error_document: list[ErrorDocumentItem] | None = None


@dataclass(kw_only=True)
class ProcessedDataDocumentItem32(OrderedItem):
    data_processing_document: dict[str, Any] | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    processed_data_identifier: TStringValue | None = None


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem32]


@dataclass(kw_only=True)
class StatisticsDocumentItem:
    statistical_feature: TClass


@dataclass(kw_only=True)
class StatisticsAggregateDocument:
    statistics_document: list[StatisticsDocumentItem] | None = None


@dataclass(kw_only=True)
class TIntegerValue1:
    field_type: TClass
    value: int


TIntegerValue = int | TIntegerValue1


@dataclass(kw_only=True)
class TQuantityValueModel:
    value: float
    unit: TUnit
    has_statistic_datum_role: TStatisticDatumRole | None = None
    field_type: TClass | None = None


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue | None = None
    description: Any | None = None
    brand_name: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    device_document: list[DeviceDocumentItem] | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    data_system_instance_identifier: TStringValue | None = None
    file_name: TStringValue | None = None
    UNC_path: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    device_identifier: TStringValue | None = None
    detection_type: TStringValue | None = None
    product_manufacturer: TStringValue | None = None
    brand_name: TStringValue | None = None
    equipment_serial_number: TStringValue | None = None
    model_number: TStringValue | None = None
    firmware_version: TStringValue | None = None
    field_index: int | None = None
    detector_wavelength_setting: TQuantityValueNanometer | None = None
    detector_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_wavelength_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_bandwidth_setting: TQuantityValueNanometer | None = None
    electronic_absorbance_reference_bandwidth_setting: TQuantityValueNanometer | None = (
        None
    )
    electronic_absorbance_reference_wavelength_setting: TQuantityValueNanometer | None = (
        None
    )
    excitation_wavelength_setting: TQuantityValueNanometer | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue
    description: Any | None = None
    batch_identifier: TStringValue | None = None
    sample_role_type: TClass | None = None
    written_name: TStringValue | None = None


@dataclass(kw_only=True)
class DataProcessingDocument:
    cell_type_processing_method: TStringValue | None = None
    cell_density_dilution_factor: TQuantityValueUnitless | None = None
    minimum_cell_diameter_setting: TQuantityValueMicrometer | None = None
    maximum_cell_diameter_setting: TQuantityValueMicrometer | None = None


@dataclass(kw_only=True)
class AnalyteDocument:
    analyte_name: TStringValue
    volume_concentration: TQuantityValueMilliliterPerLiter
    molar_concentration: TQuantityValueMillimolePerLiter
    mass_concentration: TQuantityValueGramPerLiter


@dataclass(kw_only=True)
class AnalyteDocument1:
    analyte_name: TStringValue
    molar_concentration: TQuantityValueMillimolePerLiter
    mass_concentration: TQuantityValueGramPerLiter
    volume_concentration: TQuantityValueMilliliterPerLiter | None = None


@dataclass(kw_only=True)
class AnalyteAggregateDocument:
    analyte_document: list[
        AnalyteDocument
        | AnalyteDocument1
        | AnalyteDocument10
        | AnalyteDocument100
        | AnalyteDocument102
        | AnalyteDocument103
        | AnalyteDocument104
    ]


@dataclass(kw_only=True)
class AnalyteDocument10:
    analyte_name: TStringValue
    mass_concentration: TQuantityValueGramPerLiter
    volume_concentration: TQuantityValueMilliliterPerLiter | None = None
    molar_concentration: TQuantityValueMillimolePerLiter | None = None


@dataclass(kw_only=True)
class AnalyteDocument100:
    analyte_name: TStringValue
    volume_concentration: TQuantityValueMilliliterPerLiter
    mass_concentration: TQuantityValueGramPerLiter
    molar_concentration: TQuantityValueMillimolePerLiter | None = None


@dataclass(kw_only=True)
class AnalyteDocument102:
    analyte_name: TStringValue
    volume_concentration: TQuantityValueMilliliterPerLiter
    molar_concentration: TQuantityValueMillimolePerLiter
    mass_concentration: TQuantityValueGramPerLiter | None = None


@dataclass(kw_only=True)
class AnalyteDocument103:
    analyte_name: TStringValue
    molar_concentration: TQuantityValueMillimolePerLiter
    volume_concentration: TQuantityValueMilliliterPerLiter | None = None
    mass_concentration: TQuantityValueGramPerLiter | None = None


@dataclass(kw_only=True)
class AnalyteDocument104:
    analyte_name: TStringValue
    volume_concentration: TQuantityValueMilliliterPerLiter
    molar_concentration: TQuantityValueMillimolePerLiter | None = None
    mass_concentration: TQuantityValueGramPerLiter | None = None


@dataclass(kw_only=True)
class CalculatedDataDocumentItem:
    calculated_data_name: TStringValue
    calculated_result: TQuantityValueModel
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    calculated_data_identifier: TStringValue | None = None
    calculation_description: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class CalculatedDataAggregateDocument:
    calculated_data_document: list[CalculatedDataDocumentItem]


@dataclass(kw_only=True)
class ImageDocumentItem:
    index: TIntegerValue | None = None
    experimental_data_identifier: TStringValue | None = None


@dataclass(kw_only=True)
class ImageAggregateDocument:
    image_document: list[ImageDocumentItem] | None = None


@dataclass(kw_only=True)
class MeasurementDocument16:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument17:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument18:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument19:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument20:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument21:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument22:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument23:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument24:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument25:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument26:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument27:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument28:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None


@dataclass(kw_only=True)
class MeasurementDocument29:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument30:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument31:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument48:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument49:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument50:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument51:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument52:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument53:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument54:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument55:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument56:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument57:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument58:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument59:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument60:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None


@dataclass(kw_only=True)
class MeasurementDocument61:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    absorbance: TQuantityValueMilliAbsorbanceUnit
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument62:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    detection_type: TStringValue | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class ProcessedDataDocumentItem:
    viability__cell_counter_: TQuantityValuePercent
    viable_cell_density__cell_counter_: TQuantityValueMillionCellsPerMilliliter
    data_processing_document: DataProcessingDocument | None = None
    data_source_aggregate_document: DataSourceAggregateDocument | None = None
    processed_data_identifier: TStringValue | None = None
    field_index: int | None = None
    total_cell_density__cell_counter_: TQuantityValueMillionCellsPerMilliliter | None = (
        None
    )
    dead_cell_density__cell_counter_: TQuantityValueMillionCellsPerMilliliter | None = (
        None
    )
    average_total_cell_diameter: TQuantityValueMicrometer | None = None
    average_live_cell_diameter__cell_counter_: TQuantityValueMicrometer | None = None
    average_dead_cell_diameter__cell_counter_: TQuantityValueMicrometer | None = None
    total_cell_diameter_distribution: TDatacube | None = None
    total_cell_count: TQuantityValueCell | None = None
    viable_cell_count: TQuantityValueCell | None = None
    dead_cell_count: TQuantityValueCell | None = None
    average_total_cell_circularity: TQuantityValueUnitless | None = None
    average_viable_cell_circularity: TQuantityValueUnitless | None = None


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument1:
    processed_data_document: list[ProcessedDataDocumentItem]


@dataclass(kw_only=True)
class MeasurementDocument:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument1:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument2:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument3:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument4:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument5:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument6:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument7:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument8:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument9:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument10:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument11:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    pH: TQuantityValuePH
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument12:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None


@dataclass(kw_only=True)
class MeasurementDocument13:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument14:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument15:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    analyte_aggregate_document: AnalyteAggregateDocument
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument32:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument33:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument34:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument35:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    pH: TQuantityValuePH
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument36:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument37:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument38:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None


@dataclass(kw_only=True)
class MeasurementDocument39:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    pO2: TQuantityValueMillimeterOfMercury
    pCO2: TQuantityValueMillimeterOfMercury
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    carbon_dioxide_saturation: TQuantityValuePercent | None = None
    oxygen_saturation: TQuantityValuePercent | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument40:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument41:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument42:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    temperature: TQuantityValueDegreeCelsius | None = None


@dataclass(kw_only=True)
class MeasurementDocument43:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    pH: TQuantityValuePH
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    temperature: TQuantityValueDegreeCelsius | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument44:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    absorbance: TQuantityValueMilliAbsorbanceUnit
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None


@dataclass(kw_only=True)
class MeasurementDocument45:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    absorbance: TQuantityValueMilliAbsorbanceUnit
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    mass_concentration: TQuantityValuePicogramPerMilliliter | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument46:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    osmolality: TQuantityValueMilliOsmolesPerKilogram
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementDocument47:
    measurement_time: TDateTimeStampValue
    measurement_identifier: TStringValue
    device_control_aggregate_document: DeviceControlAggregateDocument
    sample_document: SampleDocument
    processed_data_aggregate_document: ProcessedDataAggregateDocument1
    detection_type: TStringValue | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[
        MeasurementDocument
        | MeasurementDocument1
        | MeasurementDocument10
        | MeasurementDocument11
        | MeasurementDocument12
        | MeasurementDocument13
        | MeasurementDocument14
        | MeasurementDocument15
        | MeasurementDocument16
        | MeasurementDocument17
        | MeasurementDocument18
        | MeasurementDocument19
        | MeasurementDocument2
        | MeasurementDocument20
        | MeasurementDocument21
        | MeasurementDocument22
        | MeasurementDocument23
        | MeasurementDocument24
        | MeasurementDocument25
        | MeasurementDocument26
        | MeasurementDocument27
        | MeasurementDocument28
        | MeasurementDocument29
        | MeasurementDocument3
        | MeasurementDocument30
        | MeasurementDocument31
        | MeasurementDocument32
        | MeasurementDocument33
        | MeasurementDocument34
        | MeasurementDocument35
        | MeasurementDocument36
        | MeasurementDocument37
        | MeasurementDocument38
        | MeasurementDocument39
        | MeasurementDocument4
        | MeasurementDocument40
        | MeasurementDocument41
        | MeasurementDocument42
        | MeasurementDocument43
        | MeasurementDocument44
        | MeasurementDocument45
        | MeasurementDocument46
        | MeasurementDocument47
        | MeasurementDocument48
        | MeasurementDocument49
        | MeasurementDocument5
        | MeasurementDocument50
        | MeasurementDocument51
        | MeasurementDocument52
        | MeasurementDocument53
        | MeasurementDocument54
        | MeasurementDocument55
        | MeasurementDocument56
        | MeasurementDocument57
        | MeasurementDocument58
        | MeasurementDocument59
        | MeasurementDocument6
        | MeasurementDocument60
        | MeasurementDocument61
        | MeasurementDocument62
        | MeasurementDocument7
        | MeasurementDocument8
        | MeasurementDocument9
    ]
    diagnostic_trace_aggregate_document: DiagnosticTraceAggregateDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    image_aggregate_document: ImageAggregateDocument | None = None
    error_aggregate_document: ErrorAggregateDocument | None = None
    data_processing_time: TDateTimeStampValue | None = None


@dataclass(kw_only=True)
class SolutionAnalyzerDocumentItem:
    measurement_aggregate_document: MeasurementAggregateDocument
    analyst: TStringValue | None = None
    submitter: TStringValue | None = None


@dataclass(kw_only=True)
class SolutionAnalyzerAggregateDocument:
    solution_analyzer_document: list[SolutionAnalyzerDocumentItem]
    device_system_document: DeviceSystemDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    calculated_data_aggregate_document: CalculatedDataAggregateDocument | None = None
    statistics_aggregate_document: StatisticsAggregateDocument | None = None
    data_system_document: DataSystemDocument | None = None


@dataclass(kw_only=True)
class Model:
    field_asm_manifest: AdmCoreREC202403ManifestSchema | str
    solution_analyzer_aggregate_document: SolutionAnalyzerAggregateDocument | None = (
        None
    )



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import InvalidJsonFloat
from allotropy.parsers.mabtech_apex.mabtech_apex_contents import MabtechApexContents
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_from_series_or_nan,
    try_float_from_series_or_none,
    try_str_from_series,
    try_str_from_series_or_none,
)

IMAGE_FEATURES = [
    "Spot Forming Units (SFU)",
    "Average Relative Spot Volume (RSV)",
    "Sum of Spot Volume (RSV)",
]


@dataclass(frozen=True)
class PlateInformation:
    unc_path: str | None
    analyst: str | None
    software_version: str | None
    model_number: str
    equipment_serial_number: str | None

    @staticmethod
    def create(contents: MabtechApexContents) -> PlateInformation:
        plateinformation = contents.plate_info
        machine_id = assert_not_none(
            re.match(
                "([A-Z]+[a-z]+) ([0-9]+)",
                try_str_from_series(plateinformation, key="Machine ID:"),
            ),
            msg="Unable to interpret Machine ID",
        )

        return PlateInformation(
            unc_path=try_str_from_series_or_none(plateinformation, key="Path:"),
            analyst=try_str_from_series_or_none(plateinformation, key="Saved By:"),
            software_version=try_str_from_series_or_none(
                plateinformation, key="Software Version:"
            ),
            model_number=machine_id.group(1),
            equipment_serial_number=machine_id.group(2),
        )


@dataclass(frozen=True)
class Well:
    measurement_time: str
    sample_identifier: str
    location_identifier: str
    well_plate_identifier: str | None
    exposure_duration_setting: float | None
    illumination_setting: float | None
    image_features: list[dict[str, float | InvalidJsonFloat]]

    @staticmethod
    def create(plate_data: pd.Series[str]) -> Well | None:
        # if Read Date is not present in file, return None, no measurement for given Well
        if try_str_from_series_or_none(plate_data, "Read Date") is None:
            return None

        location_id = try_str_from_series(plate_data, "Well")

        well_plate = try_str_from_series_or_none(plate_data, "Plate")

        image_feature_list = []
        for feature in IMAGE_FEATURES:
            image_dict = {
                str(feature): try_float_from_series_or_nan(plate_data, feature)
            }
            image_feature_list.append(image_dict)

        return Well(
            measurement_time=try_str_from_series(plate_data, "Read Date"),
            location_identifier=location_id,
            well_plate_identifier=well_plate,
            sample_identifier=f"{well_plate}_{location_id}",
            exposure_duration_setting=try_float_from_series_or_none(
                plate_data, "Exposure"
            ),
            illumination_setting=try_float_from_series_or_none(
                plate_data, "Preset Intensity"
            ),
            image_features=image_feature_list,
        )


@dataclass(frozen=True)
class WellList:
    wells: list[Well]

    def __iter__(self) -> Iterator[Well]:
        return iter(self.wells)

    @staticmethod
    def create(contents: MabtechApexContents) -> WellList:
        plate_data = contents.data
        well_list = []
        for _, well_data in plate_data.iterrows():
            well = Well.create(well_data)
            if well is not None:
                well_list.append(well)
        return WellList(wells=well_list)



# mypy: disallow_any_generics = False

from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator
from dataclasses import dataclass
import re

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    TransmittedLightSetting,
)
from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader
from allotropy.parsers.agilent_gen5_image.constants import (
    AUTOFOCUS_STRINGS,
    CHANNEL_HEADER_REGEX,
    DEFAULT_EXPORT_FORMAT_ERROR,
    DetectionType,
    DETECTOR_DISTANCE_REGEX,
    FILENAME_REGEX,
    HEADER_PREFIXES,
    ReadType,
    SETTINGS_SECTION_REGEX,
    TRANSMITTED_LIGHT_MAP,
    UNSUPORTED_READ_TYPE_ERROR,
)
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_float_or_nan,
    try_float_or_none,
)


def read_data_section(reader: LinesReader) -> str:
    data_section = "\n".join(
        [
            reader.pop() or "",
            *reader.pop_until_empty(),
        ]
    )
    reader.drop_empty()
    return data_section


def parse_settings(settings: list[str]) -> dict:
    """Returns a dictionary containing all key values identified in a list of settings.

    If there are additional (non key value) settings, they are also returned under the
    'non_kv_settings' key.

    Supported settings lines format:
        - key: value -> returned as {'key': 'value}
        - key1: value1, key2: value2, ... -> returned as {'key1': 'value1', key2', 'value2', ...}
        - Non keyvalue setting -> returned as {'non_kv_settings': ['Non keyvalue setting', ...]}
    """
    settings_dict: dict = {"non_kv_settings": []}
    for line in settings:
        strp_line = str(line.strip())

        line_data: list[str] = strp_line.split(", ")
        for read_datum in line_data:
            splitted_datum = read_datum.split(": ")
            if len(splitted_datum) == 1:
                settings_dict["non_kv_settings"].append(splitted_datum[0])
            elif len(splitted_datum) == 2:  # noqa: PLR2004
                settings_dict[splitted_datum[0]] = splitted_datum[1]

    return settings_dict


@dataclass(frozen=True)
class HeaderData:
    software_version: str
    experiment_file_path: str
    protocol_file_path: str
    datetime: str
    well_plate_identifier: str
    model_number: str
    equipment_serial_number: str

    @classmethod
    def create(cls, reader: LinesReader, file_name: str) -> HeaderData:
        assert_not_none(reader.drop_until("^Software Version"), "Software Version")
        metadata_dict = cls._parse_metadata(reader)
        datetime_ = cls._parse_datetime(metadata_dict["Date"], metadata_dict["Time"])
        plate_identifier = cls._get_identifier_from_filename_or_none(file_name)

        return HeaderData(
            software_version=metadata_dict["Software Version"],
            experiment_file_path=metadata_dict["Experiment File Path:"],
            protocol_file_path=metadata_dict["Protocol File Path:"],
            datetime=datetime_,
            well_plate_identifier=plate_identifier or metadata_dict["Plate Number"],
            model_number=metadata_dict["Reader Type:"],
            equipment_serial_number=metadata_dict["Reader Serial Number:"],
        )

    @classmethod
    def _parse_metadata(cls, reader: LinesReader) -> dict:
        metadata_dict: dict = {}
        for metadata_line in reader.pop_until("Procedure Details"):
            reader.drop_empty()
            line_split = metadata_line.split("\t")
            if line_split[0] not in HEADER_PREFIXES:
                msg = msg_for_error_on_unrecognized_value(
                    "metadata key", line_split[0], HEADER_PREFIXES
                )
                raise AllotropeConversionError(msg)
            try:
                metadata_dict[line_split[0]] = line_split[1]
            except IndexError:
                metadata_dict[line_split[0]] = ""

        return metadata_dict

    @classmethod
    def _parse_datetime(cls, date_: str, time_: str) -> str:
        return f"{date_} {time_}"

    @classmethod
    def _get_identifier_from_filename_or_none(cls, file_name: str) -> str | None:
        matches = re.match(FILENAME_REGEX, file_name)
        if not matches:
            return None

        matches_dict = matches.groupdict()
        return matches_dict["plate_identifier"]


@dataclass
class InstrumentSettings:
    auto_focus: bool
    detector_distance: float | None
    fluorescent_tag: str | None = None
    excitation_wavelength: float | None = None
    detector_wavelength: float | None = None
    transmitted_light: TransmittedLightSetting | None = None
    illumination: float | None = None
    exposure_duration: float | None = None
    detector_gain: str | None = None

    @classmethod
    def create(cls, settings_lines: list[str]) -> InstrumentSettings:
        channel_settings = cls._get_channel_line_settings(settings_lines[0])

        settings_dict = parse_settings(settings_lines[1:])
        non_kv_settings = settings_dict["non_kv_settings"]

        if exposure_duration := settings_dict.get("Integration time"):
            exposure_duration = str(exposure_duration).split()[0]

        return InstrumentSettings(
            auto_focus=cls._get_auto_focus(non_kv_settings),
            detector_distance=cls._get_detector_distance(non_kv_settings),
            fluorescent_tag=channel_settings.get("fluorescent_tag"),
            excitation_wavelength=try_float_or_none(
                channel_settings.get("excitation_wavelength")
            ),
            detector_wavelength=try_float_or_none(
                channel_settings.get("detector_wavelength")
            ),
            transmitted_light=cls._get_transmitted_light(non_kv_settings),
            illumination=try_float_or_none(settings_dict.get("LED intensity")),
            exposure_duration=try_float_or_none(exposure_duration),
            detector_gain=settings_dict.get("Camera gain"),
        )

    @classmethod
    def _get_channel_line_settings(cls, settings_header: str) -> dict:
        if matches := re.match(CHANNEL_HEADER_REGEX, settings_header):
            return matches.groupdict()
        return {}

    @classmethod
    def _get_auto_focus(cls, settings: list[str]) -> bool:
        return any(str_ in settings for str_ in AUTOFOCUS_STRINGS)

    @classmethod
    def _get_detector_distance(cls, settings: list[str]) -> float | None:
        if match := re.search(DETECTOR_DISTANCE_REGEX, "\n".join(settings)):
            return try_float(match.groups()[0], "Detector Distance")
        return None

    @classmethod
    def _get_transmitted_light(
        cls, settings: list[str]
    ) -> TransmittedLightSetting | None:
        for line in settings:
            if line in TRANSMITTED_LIGHT_MAP:
                return TRANSMITTED_LIGHT_MAP[line]
        return None


@dataclass(frozen=True)
class ReadSection:
    image_mode: DetectionType
    magnification_setting: float | None
    image_count_setting: float | None
    instrument_settings_list: list[InstrumentSettings]

    @classmethod
    def create(cls, reader: LinesReader) -> ReadSection:
        top_read_lines = list(reader.pop_until(SETTINGS_SECTION_REGEX))

        detection_type = cls._get_detection_type("\n".join(top_read_lines))
        instrument_settings_list = cls._get_instrument_settings_list(reader)

        bottom_read_lines = list(reader.pop_until_empty())
        read_settings = parse_settings(top_read_lines + bottom_read_lines)

        if objective := read_settings.get("Objective"):
            objective = str(objective).replace("x", "")

        return ReadSection(
            image_mode=detection_type,
            magnification_setting=try_float_or_none(objective),
            image_count_setting=cls._get_image_count_setting(read_settings),
            instrument_settings_list=instrument_settings_list,
        )

    @classmethod
    def _get_detection_type(cls, read_chunk: str) -> DetectionType:
        if DetectionType.SINGLE_IMAGE.value in read_chunk:
            return DetectionType.SINGLE_IMAGE
        elif DetectionType.MONTAGE.value in read_chunk:
            return DetectionType.MONTAGE
        elif DetectionType.Z_STACK.value in read_chunk:
            return DetectionType.Z_STACK

        msg = f"Measurement mode not found; expected to find one of {sorted(DetectionType._member_names_)}."

        raise AllotropeConversionError(msg)

    @classmethod
    def _get_settings_sections(cls, reader: LinesReader) -> Iterator[list[str]]:
        while True:
            initial_line = reader.get() or ""
            if not re.search(SETTINGS_SECTION_REGEX, initial_line):
                break
            reader.pop()
            lines = [initial_line, *reader.pop_until(r"^\t\w")]
            yield lines

    @classmethod
    def _get_instrument_settings_list(
        cls, reader: LinesReader
    ) -> list[InstrumentSettings]:
        settings_list = []
        auto_focus = False
        for idx, settings_section in enumerate(cls._get_settings_sections(reader)):
            instrument_settings = InstrumentSettings.create(settings_section)
            # The autofocus setting is only reported in the first section but applies to all others
            if idx == 0:
                auto_focus = instrument_settings.auto_focus
            else:
                instrument_settings.auto_focus = auto_focus
            settings_list.append(instrument_settings)
        return settings_list

    @classmethod
    def _get_image_count_setting(cls, read_settings: dict) -> float | None:
        montage_rows = read_settings.get("Montage rows")
        montage_columns = read_settings.get("columns")
        if montage_rows and montage_columns:
            return try_float(montage_rows, "Montage rows") * try_float(
                montage_columns, "Montage columns"
            )
        return None


@dataclass(frozen=True)
class ReadData:
    read_sections: list[ReadSection]

    @classmethod
    def create(cls, reader: LinesReader) -> ReadData:
        assert_not_none(reader.drop_until("^Procedure Details"), "Procedure Details")
        reader.pop()
        reader.drop_empty()
        procedure_details = read_data_section(reader)

        read_type = cls._get_read_type(procedure_details)
        if read_type != ReadType.IMAGE:
            raise AllotropeConversionError(UNSUPORTED_READ_TYPE_ERROR)

        section_lines_reader = SectionLinesReader(procedure_details.splitlines())

        return ReadData(
            read_sections=[
                ReadSection.create(read_section)
                for read_section in section_lines_reader.iter_sections("^Read\t")
            ]
        )

    @classmethod
    def _get_read_type(cls, procedure_details: str) -> ReadType:
        if ReadType.KINETIC.value in procedure_details:
            return ReadType.KINETIC
        elif ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            return ReadType.SPECTRAL
        elif ReadType.ENDPOINT.value in procedure_details:
            return ReadType.ENDPOINT
        elif ReadType.IMAGE.value in procedure_details:
            return ReadType.IMAGE

        msg = f"Read type not found; expected to find one of {sorted(ReadType._member_names_)}."
        raise AllotropeConversionError(msg)


@dataclass(frozen=True)
class LayoutData:
    sample_identifiers: dict

    @staticmethod
    def create_default() -> LayoutData:
        return LayoutData(sample_identifiers={})

    @staticmethod
    def create(layout_str: str) -> LayoutData:
        identifiers = {}

        layout_lines: list[str] = layout_str.splitlines()
        # first line is "Layout", second line is column numbers
        current_row = "A"
        for i in range(2, len(layout_lines)):
            split_line = layout_lines[i].split("\t")
            if split_line[0]:
                current_row = split_line[0]
            label = split_line[-1]
            for j in range(1, len(split_line) - 1):
                well_loc = f"{current_row}{j}"
                if label == "Name":
                    identifiers[well_loc] = split_line[j]
                elif label == "Well ID":
                    # give prevalence to the "Name" field if present
                    sample_id = identifiers.get(well_loc)
                    identifiers[well_loc] = sample_id if sample_id else split_line[j]

        return LayoutData(sample_identifiers=identifiers)


@dataclass(frozen=True)
class ActualTemperature:
    value: float | None = None

    @staticmethod
    def create_default() -> ActualTemperature:
        return ActualTemperature()

    @staticmethod
    def create(actual_temperature: str) -> ActualTemperature:
        if len(actual_temperature.split("\n")) != 1:
            msg = f"Expected the Temperature section '{actual_temperature}' to contain exactly 1 line."
            raise AllotropeConversionError(msg)

        return ActualTemperature(
            value=try_float(
                actual_temperature.strip().split("\t")[-1], "Actual Temperature"
            ),
        )


@dataclass(frozen=True)
class ImageFeature:
    identifier: str
    name: str
    result: JsonFloat


@dataclass(frozen=True)
class Results:
    image_features: dict[str, list[ImageFeature]]

    @staticmethod
    def create(results: str) -> Results:
        image_features = defaultdict(list)

        result_lines = results.splitlines()

        if result_lines[0].strip() != "Results":
            msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
            raise AllotropeConversionError(msg)

        # result_lines[1] contains column numbers
        current_row = "A"
        for row_num in range(2, len(result_lines)):
            values = result_lines[row_num].split("\t")
            if values[0]:
                current_row = values[0]
            feature_name = values[-1]
            for col_num in range(1, len(values) - 1):
                well_position = f"{current_row}{col_num}"
                well_value = try_float_or_nan(values[col_num])

                image_features[well_position].append(
                    ImageFeature(
                        identifier=random_uuid_str(),
                        name=feature_name,
                        result=well_value,
                    )
                )

        return Results(image_features)


@dataclass(frozen=True)
class PlateData:
    header_data: HeaderData
    read_data: ReadData
    layout_data: LayoutData
    results: Results

    @staticmethod
    def create(reader: LinesReader, file_name: str) -> PlateData:
        header_data = HeaderData.create(reader, file_name)
        read_data = ReadData.create(reader)
        layout_data = LayoutData.create_default()
        results = None

        while reader.current_line_exists():
            data_section = read_data_section(reader)
            if data_section.startswith("Layout"):
                layout_data = LayoutData.create(data_section)
            elif data_section.startswith("Results"):
                results = Results.create(data_section)

        # If there is no results table, it might mean that the export format includes results in
        # separate tables, or that there are no results at all (also bad format)
        if results is None:
            raise AllotropeConversionError(DEFAULT_EXPORT_FORMAT_ERROR)

        return PlateData(
            header_data=header_data,
            read_data=read_data,
            layout_data=layout_data,
            results=results,
        )



# mypy: disallow_any_generics = False

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
import re

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.parsers.agilent_gen5.constants import (
    EMISSION_KEY,
    EXCITATION_KEY,
    FILENAME_REGEX,
    GAIN_KEY,
    HEADER_PREFIXES,
    MEASUREMENTS_DATA_POINT_KEY,
    MIRROR_KEY,
    OPTICS_KEY,
    PATHLENGTH_CORRECTION_KEY,
    READ_HEIGHT_KEY,
    READ_SPEED_KEY,
    ReadMode,
    ReadType,
    UNSUPORTED_READ_TYPE_ERROR,
    WAVELENGTHS_KEY,
)
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_float_or_nan,
    try_float_or_none,
)


def read_data_section(reader: LinesReader) -> str:
    data_section = "\n".join(
        [
            reader.pop() or "",
            *reader.pop_until_empty(),
        ]
    )
    reader.drop_empty()
    return data_section


@dataclass(frozen=True)
class HeaderData:
    software_version: str
    experiment_file_path: str
    protocol_file_path: str
    datetime: str
    well_plate_identifier: str
    model_number: str
    equipment_serial_number: str

    @classmethod
    def create(cls, reader: LinesReader, file_name: str) -> HeaderData:
        assert_not_none(reader.drop_until("^Software Version"), "Software Version")
        metadata_dict = cls._parse_metadata(reader)
        datetime_ = cls._parse_datetime(metadata_dict["Date"], metadata_dict["Time"])
        plate_identifier = cls._get_identifier_from_filename_or_none(file_name)

        return HeaderData(
            software_version=metadata_dict["Software Version"],
            experiment_file_path=metadata_dict["Experiment File Path:"],
            protocol_file_path=metadata_dict["Protocol File Path:"],
            datetime=datetime_,
            well_plate_identifier=plate_identifier or metadata_dict["Plate Number"],
            model_number=metadata_dict["Reader Type:"],
            equipment_serial_number=metadata_dict["Reader Serial Number:"],
        )

    @classmethod
    def _parse_metadata(cls, reader: LinesReader) -> dict:
        metadata_dict: dict = {}
        for metadata_line in reader.pop_until("Procedure Details"):
            reader.drop_empty()
            line_split = metadata_line.split("\t")
            if line_split[0] not in HEADER_PREFIXES:
                msg = msg_for_error_on_unrecognized_value(
                    "metadata key", line_split[0], HEADER_PREFIXES
                )
                raise AllotropeConversionError(msg)
            try:
                metadata_dict[line_split[0]] = line_split[1]
            except IndexError:
                metadata_dict[line_split[0]] = ""

        return metadata_dict

    @classmethod
    def _parse_datetime(cls, date_: str, time_: str) -> str:
        return f"{date_} {time_}"

    @classmethod
    def _get_identifier_from_filename_or_none(cls, file_name: str) -> str | None:
        matches = re.match(FILENAME_REGEX, file_name)
        if not matches:
            return None

        matches_dict = matches.groupdict()
        return matches_dict["plate_identifier"]


@dataclass(frozen=True)
class FilterSet:
    gain: str
    emission: str | None = None
    excitation: str | None = None
    mirror: str | None = None
    optics: str | None = None

    @property
    def detector_wavelength_setting(self) -> float | None:
        if self.emission == "Full light" or not self.emission:
            return None
        return try_float(self.emission.split("/")[0], "Detector wavelength")

    @property
    def detector_bandwidth_setting(self) -> float | None:
        if not self.emission or self.emission == "Full light":
            return None
        try:
            return try_float(self.emission.split("/")[1], "Detector bandwith")
        except IndexError:
            return None

    @property
    def excitation_wavelength_setting(self) -> float | None:
        if self.excitation:
            return try_float(self.excitation.split("/")[0], "Excitation wavelength")
        return None

    @property
    def excitation_bandwidth_setting(self) -> float | None:
        if not self.excitation:
            return None
        try:
            return try_float(self.excitation.split("/")[1], "Excitation bandwith")
        except IndexError:
            return None

    @property
    def wavelength_filter_cutoff_setting(self) -> float | None:
        if self.mirror:
            return try_float(self.mirror.split(" ")[1], "Wavelength filter cutoff")
        return None

    @property
    def scan_position_setting(self) -> ScanPositionSettingPlateReader | None:
        position_lookup = {
            "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
            "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
        }
        position = self.optics
        if self.mirror:
            position = self.mirror.split(" ")[0]

        # TODO: Raise if position is not a valid lookup value
        return position_lookup.get(position) if position is not None else None


@dataclass(frozen=True)
class ReadData:
    read_mode: ReadMode
    measurement_labels: list[str]
    pathlength_correction: str | None
    step_label: str | None
    number_of_averages: float | None
    detector_distance: float | None
    detector_carriage_speed: str | None
    filter_sets: dict[str, FilterSet]

    @classmethod
    def create(cls, reader: LinesReader) -> ReadData:
        assert_not_none(reader.drop_until("^Procedure Details"), "Procedure Details")
        reader.pop()
        reader.drop_empty()
        procedure_details = read_data_section(reader)

        read_type = cls.get_read_type(procedure_details)
        if read_type != ReadType.ENDPOINT:
            raise AllotropeConversionError(UNSUPORTED_READ_TYPE_ERROR)

        read_mode = cls.get_read_mode(procedure_details)
        device_control_data = cls._get_device_control_data(procedure_details, read_mode)
        measurement_labels = cls._get_measurement_labels(device_control_data, read_mode)

        number_of_averages = device_control_data.get(MEASUREMENTS_DATA_POINT_KEY)
        read_height = device_control_data.get(READ_HEIGHT_KEY, "")

        return ReadData(
            read_mode=read_mode,
            step_label=device_control_data.get("Step Label"),
            measurement_labels=measurement_labels,
            detector_carriage_speed=device_control_data.get(READ_SPEED_KEY),
            # Absorbance attributes
            pathlength_correction=device_control_data.get(PATHLENGTH_CORRECTION_KEY),
            number_of_averages=try_float_or_none(number_of_averages),
            # Luminescence attributes
            detector_distance=try_float_or_none(read_height.split(" ")[0]),
            # Fluorescence attributes
            filter_sets=cls._get_filter_sets(
                measurement_labels, device_control_data, read_mode
            ),
        )

    @staticmethod
    def get_read_mode(procedure_details: str) -> ReadMode:
        if ReadMode.ABSORBANCE.value in procedure_details:
            return ReadMode.ABSORBANCE
        elif (
            ReadMode.FLUORESCENCE.value in procedure_details
            or ReadMode.ALPHALISA.value in procedure_details
        ):
            return ReadMode.FLUORESCENCE
        elif ReadMode.LUMINESCENCE.value in procedure_details:
            return ReadMode.LUMINESCENCE

        msg = f"Read mode not found; expected to find one of {sorted(ReadMode._member_names_)}."
        raise AllotropeConversionError(msg)

    @staticmethod
    def get_read_type(procedure_details: str) -> ReadType:
        if ReadType.KINETIC.value in procedure_details:
            return ReadType.KINETIC
        elif ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            return ReadType.SPECTRAL
        # check for this last, because other modes still contain the word "Endpoint"
        elif ReadType.ENDPOINT.value in procedure_details:
            return ReadType.ENDPOINT

        msg = f"Read type not found; expected to find one of {sorted(ReadType._member_names_)}."
        raise AllotropeConversionError(msg)

    @classmethod
    def _get_measurement_labels(cls, device_control_data: dict, read_mode: str) -> list:
        step_label = device_control_data.get("Step Label")
        label_prefix = f"{step_label}:" if step_label else ""
        measurement_labels = []

        if read_mode == ReadMode.ABSORBANCE:
            measurement_labels = cls._get_absorbance_measurement_labels(
                label_prefix, device_control_data
            )

        if read_mode == ReadMode.FLUORESCENCE:
            excitations = device_control_data.get(EXCITATION_KEY, [])
            emissions = device_control_data.get(EMISSION_KEY, [])
            measurement_labels = [
                f"{label_prefix}{excitation},{emission}"
                for excitation, emission in zip(excitations, emissions, strict=True)
            ]
            if not measurement_labels:
                measurement_labels = ["Alpha"]

        if read_mode == ReadMode.LUMINESCENCE:
            emissions = device_control_data.get(EMISSION_KEY)
            for emission in emissions:
                label = "Lum" if emission == "Full light" else emission
                measurement_labels.append(f"{label_prefix}{label}")

        return measurement_labels

    @classmethod
    def _get_absorbance_measurement_labels(
        cls, label_prefix: str | None, device_control_data: dict
    ) -> list:
        wavelengths = device_control_data.get(WAVELENGTHS_KEY, [])
        pathlength_correction = device_control_data.get(PATHLENGTH_CORRECTION_KEY)
        measurement_labels = []

        for wavelenght in wavelengths:
            label = f"{label_prefix}{wavelenght}"
            measurement_labels.append(label)

        if pathlength_correction:
            test, ref = pathlength_correction.split(" / ")
            test_label = f"{label_prefix}{test} [Test]"
            ref_label = f"{label_prefix}{ref} [Ref]"
            measurement_labels.extend([test_label, ref_label])

        return measurement_labels

    @classmethod
    def _get_device_control_data(
        cls, procedure_details: str, read_mode: ReadMode
    ) -> dict:
        list_keys = frozenset(
            {
                EMISSION_KEY,
                EXCITATION_KEY,
                OPTICS_KEY,
                GAIN_KEY,
                MIRROR_KEY,
                WAVELENGTHS_KEY,
            }
        )
        read_data_dict: dict = {label: [] for label in list_keys}
        read_lines: list[str] = procedure_details.splitlines()

        for line in read_lines:
            strp_line = str(line.strip())
            if strp_line.startswith("Read\t"):
                read_data_dict["Step Label"] = cls._get_step_label(line, read_mode)
                continue

            elif strp_line.startswith(WAVELENGTHS_KEY):
                wavelengths = strp_line.split(":  ")
                read_data_dict[WAVELENGTHS_KEY].extend(wavelengths[1].split(", "))
                continue

            line_data: list[str] = strp_line.split(",  ")
            for read_datum in line_data:
                splitted_datum = read_datum.split(": ")
                if len(splitted_datum) != 2:  # noqa: PLR2004
                    continue
                if splitted_datum[0] in list_keys:
                    read_data_dict[splitted_datum[0]].append(splitted_datum[1])
                else:
                    read_data_dict[splitted_datum[0]] = splitted_datum[1]

        return read_data_dict

    @classmethod
    def _get_step_label(cls, read_line: str, read_mode: str) -> str | None:
        split_line = read_line.split("\t")
        if len(split_line) != 2:  # noqa: PLR2004
            msg = (
                f"Expected the Read data line {split_line} to contain exactly 2 values."
            )
            raise AllotropeConversionError(msg)
        if split_line[1] != f"{read_mode.title()} Endpoint":
            return split_line[1]

        return None

    @classmethod
    def _get_filter_sets(
        cls,
        measurement_labels: list[str],
        device_control_data: dict,
        read_mode: ReadMode,
    ) -> dict[str, FilterSet]:
        filter_data: dict[str, FilterSet] = {}
        if read_mode == ReadMode.ABSORBANCE:
            return filter_data

        emissions = device_control_data.get(EMISSION_KEY, [])
        excitations = device_control_data.get(EXCITATION_KEY, [])
        mirrors = device_control_data.get(MIRROR_KEY, [])
        optics = device_control_data.get(OPTICS_KEY, [])
        gains = device_control_data.get(GAIN_KEY, [])

        for idx, label in enumerate(measurement_labels):
            mirror = None
            if mirrors and read_mode == ReadMode.FLUORESCENCE:
                mirror = mirrors[idx]
            filter_data[label] = FilterSet(
                emission=emissions[idx] if emissions else None,
                gain=gains[idx],
                excitation=excitations[idx] if excitations else None,
                mirror=mirror,
                optics=optics[idx] if optics else None,
            )
        return filter_data


@dataclass(frozen=True)
class LayoutData:
    sample_identifiers: dict

    @staticmethod
    def create_default() -> LayoutData:
        return LayoutData(sample_identifiers={})

    @staticmethod
    def create(layout_str: str) -> LayoutData:
        identifiers = {}

        layout_lines: list[str] = layout_str.splitlines()
        # first line is "Layout", second line is column numbers
        current_row = "A"
        for i in range(2, len(layout_lines)):
            split_line = layout_lines[i].split("\t")
            if split_line[0]:
                current_row = split_line[0]
            label = split_line[-1]
            for j in range(1, len(split_line) - 1):
                well_loc = f"{current_row}{j}"
                if label == "Name":
                    identifiers[well_loc] = split_line[j]
                elif label == "Well ID":
                    # give prevalence to the "Name" field if present
                    sample_id = identifiers.get(well_loc)
                    identifiers[well_loc] = sample_id if sample_id else split_line[j]

        return LayoutData(sample_identifiers=identifiers)


@dataclass(frozen=True)
class ActualTemperature:
    value: float | None = None

    @staticmethod
    def create_default() -> ActualTemperature:
        return ActualTemperature()

    @staticmethod
    def create(actual_temperature: str) -> ActualTemperature:
        if len(actual_temperature.split("\n")) != 1:
            msg = f"Expected the Temperature section '{actual_temperature}' to contain exactly 1 line."
            raise AllotropeConversionError(msg)

        return ActualTemperature(
            value=try_float(
                actual_temperature.strip().split("\t")[-1], "Actual Temperature"
            ),
        )


@dataclass(frozen=True)
class Measurement:
    identifier: str
    value: JsonFloat
    label: str


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: ReadMode


@dataclass(frozen=True)
class CalculatedDatum:
    identifier: str
    data_sources: list[DataSource]
    name: str
    result: JsonFloat


@dataclass(frozen=True)
class Results:
    measurements: defaultdict[str, list[Measurement]]
    calculated_data: list[CalculatedDatum]
    wells: list

    @staticmethod
    def create() -> Results:
        return Results(
            measurements=defaultdict(list),
            calculated_data=[],
            wells=[],
        )

    def parse_results(
        self,
        results: str,
        read_data: ReadData,
    ) -> None:
        result_lines = results.splitlines()
        calculated_data: defaultdict[str, list] = defaultdict(list)
        if result_lines[0].strip() != "Results":
            msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
            raise AllotropeConversionError(msg)

        # result_lines[1] contains column numbers
        current_row = "A"
        for row_num in range(2, len(result_lines)):
            values = result_lines[row_num].split("\t")
            if values[0]:
                current_row = values[0]
            label = values[-1]  # last column gives information about the type of read
            for col_num in range(1, len(values) - 1):
                well_pos = f"{current_row}{col_num}"
                if well_pos not in self.wells:
                    self.wells.append(well_pos)
                well_value = try_float_or_nan(values[col_num])

                if label in read_data.measurement_labels:
                    self.measurements[well_pos].append(
                        Measurement(random_uuid_str(), well_value, label)
                    )
                else:
                    calculated_data[well_pos].append([label, well_value])

        for well in self.wells:
            measurements = self.measurements[well]
            for label, value in calculated_data[well]:
                sources = self._get_sources_for_calculated_data(measurements, label)
                self.calculated_data.append(
                    CalculatedDatum(
                        identifier=random_uuid_str(),
                        data_sources=[
                            DataSource(
                                identifier=measurement.identifier,
                                feature=read_data.read_mode,
                            )
                            for measurement in sources
                        ],
                        name=label,
                        result=value,
                    )
                )

    def _get_sources_for_calculated_data(
        self, measurements: list[Measurement], calculated_data_label: str
    ) -> list[Measurement]:

        # Pathlength is a special case, its sources are always determined
        # by the pathlength correction setting
        if calculated_data_label.split(":")[-1] == "Pathlength":
            pathlength_suffixes = ["[Test]", "[Ref]"]
            sources = [
                measurement
                for measurement in measurements
                if measurement.label.split(" ")[-1] in pathlength_suffixes
            ]
        else:
            sources = [
                measurement
                for measurement in measurements
                if measurement.label in calculated_data_label
            ]

        # if there are no matches in the measurement labels, use all measurements as source
        return sources or measurements


@dataclass(frozen=True)
class PlateData:
    header_data: HeaderData
    read_data: ReadData
    layout_data: LayoutData
    results: Results
    compartment_temperature: float | None

    @staticmethod
    def create(reader: LinesReader, file_name: str) -> PlateData:
        header_data = HeaderData.create(reader, file_name)
        read_data = ReadData.create(reader)
        layout_data = LayoutData.create_default()
        actual_temperature = ActualTemperature.create_default()
        results = Results.create()

        while reader.current_line_exists():
            data_section = read_data_section(reader)
            if data_section.startswith("Layout"):
                layout_data = LayoutData.create(data_section)
            elif data_section.startswith("Actual Temperature"):
                actual_temperature = ActualTemperature.create(data_section)
            elif data_section.startswith("Results"):
                results.parse_results(
                    data_section,
                    read_data,
                )

        return PlateData(
            header_data=header_data,
            read_data=read_data,
            layout_data=layout_data,
            results=results,
            compartment_temperature=actual_temperature.value,
        )



from __future__ import annotations

from dataclasses import dataclass
from xml.etree import ElementTree as ET  # noqa: N817

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.agilent_tapestation_analysis.constants import (
    NO_SCREEN_TAPE_ID_MATCH,
    NON_CALCULATED_DATA_TAGS_PEAK,
    NON_CALCULATED_DATA_TAGS_REGION,
    NON_CALCULATED_DATA_TAGS_SAMPLE,
    UNIT_CLASS_LOOKUP,
    UNIT_CLASSES,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
    Referenceable,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    get_element_from_xml,
    get_val_from_xml,
    get_val_from_xml_or_none,
    try_float_or_nan,
    try_float_or_none,
)
from allotropy.types import IOType


def _get_calculated_data(
    element: ET.Element, excluded_tags: list[str], source_id: str, feature: str
) -> list[CalculatedDocument]:
    calculated_data = []
    for node in element:
        if (name := node.tag) in excluded_tags:
            continue
        if (value := try_float_or_none(node.text)) is None:
            continue
        calculated_data.append(
            CalculatedDocument(
                uuid=random_uuid_str(),
                name=name,
                value=value,
                data_sources=[
                    DataSource(feature=feature, reference=Referenceable(uuid=source_id))
                ],
            )
        )

    return calculated_data


@dataclass(frozen=True)
class Metadata:
    unit_cls: UNIT_CLASSES
    analyst: str | None
    analytical_method_identifier: str | None
    data_system_instance_identifier: str | None
    device_identifier: str | None
    equipment_serial_number: str | None
    experimental_data_identifier: str | None
    method_version: str | None
    software_version: str | None

    @staticmethod
    def create(root_element: ET.Element) -> Metadata:
        file_information = get_element_from_xml(root_element, "FileInformation")
        environment = get_element_from_xml(
            root_element, "ScreenTapes/ScreenTape/Environment"
        )

        return Metadata(
            unit_cls=Metadata._get_unit_class(root_element),
            analyst=get_val_from_xml_or_none(environment, "Experimenter"),
            analytical_method_identifier=get_val_from_xml_or_none(
                file_information, "Assay"
            ),
            data_system_instance_identifier=get_val_from_xml_or_none(
                environment, "Computer"
            ),
            device_identifier=get_val_from_xml_or_none(environment, "InstrumentType"),
            equipment_serial_number=get_val_from_xml_or_none(
                environment, "InstrumentSerialNumber"
            ),
            experimental_data_identifier=get_val_from_xml_or_none(
                file_information, "FileName"
            ),
            # If any, only one of those should appear, so we arbitrarily take the first one
            method_version=get_val_from_xml_or_none(file_information, "RINeVersion")
            or get_val_from_xml_or_none(file_information, "DINVersion"),
            software_version=get_val_from_xml_or_none(environment, "AnalysisVersion"),
        )

    @staticmethod
    def _get_unit_class(
        root_element: ET.Element,
    ) -> UNIT_CLASSES:
        peak_unit = get_element_from_xml(
            root_element, "Assay/Units/MolecularWeightUnit"
        ).text
        try:
            peak_unit = peak_unit or ""
            return UNIT_CLASS_LOOKUP[peak_unit]
        except KeyError as e:
            msg = f"Unrecognized Molecular Weight Unit: {peak_unit}"
            raise AllotropeConversionError(msg) from e


@dataclass(frozen=True)
class Peak:
    peak_identifier: str
    peak_height: JsonFloat
    peak_start: JsonFloat
    peak_end: JsonFloat
    peak_position: JsonFloat
    peak_area: JsonFloat
    relative_peak_area: JsonFloat
    relative_corrected_peak_area: JsonFloat
    peak_name: str | None
    comment: str | None
    calculated_data: list[CalculatedDocument]

    @staticmethod
    def create(peak_element: ET.Element) -> Peak:
        peak_height = get_val_from_xml_or_none(peak_element, "Height")
        peak_start = get_val_from_xml_or_none(peak_element, "FromMW")
        peak_end = get_val_from_xml_or_none(peak_element, "ToMW")
        peak_position = get_val_from_xml_or_none(peak_element, "Size")
        peak_area = get_val_from_xml_or_none(peak_element, "Area")
        relative_peak_area = get_val_from_xml_or_none(peak_element, "PercentOfTotal")
        relative_corrected_peak_area = get_val_from_xml_or_none(
            peak_element, "PercentIntegratedArea"
        )
        comment = get_val_from_xml_or_none(peak_element, "Comment")
        observations = get_val_from_xml_or_none(peak_element, "Observations")
        peak_identifier = random_uuid_str()

        return Peak(
            peak_identifier=peak_identifier,
            peak_height=try_float_or_nan(peak_height),
            peak_start=try_float_or_nan(peak_start),
            peak_end=try_float_or_nan(peak_end),
            peak_position=try_float_or_nan(peak_position),
            peak_area=try_float_or_nan(peak_area),
            relative_peak_area=try_float_or_nan(relative_peak_area),
            relative_corrected_peak_area=try_float_or_nan(relative_corrected_peak_area),
            peak_name=get_val_from_xml_or_none(peak_element, "Number"),
            comment=f"{comment or ''} {observations or ''}".strip() or None,
            calculated_data=_get_calculated_data(
                element=peak_element,
                excluded_tags=NON_CALCULATED_DATA_TAGS_PEAK,
                source_id=peak_identifier,
                feature="peak",
            ),
        )


@dataclass(frozen=True)
class DataRegion:
    region_identifier: str
    region_start: JsonFloat
    region_end: JsonFloat
    region_area: JsonFloat
    relative_region_area: JsonFloat
    region_name: str | None
    comment: str | None
    calculated_data: list[CalculatedDocument]

    @staticmethod
    def create(region_element: ET.Element, region_name: str) -> DataRegion:
        region_start = get_val_from_xml_or_none(region_element, "From")
        region_end = get_val_from_xml_or_none(region_element, "To")
        region_area = get_val_from_xml_or_none(region_element, "Area")
        relative_region_area = get_val_from_xml_or_none(
            region_element, "PercentOfTotal"
        )
        region_identifier = random_uuid_str()

        return DataRegion(
            region_identifier=region_identifier,
            region_start=try_float_or_nan(region_start),
            region_end=try_float_or_nan(region_end),
            region_area=try_float_or_nan(region_area),
            relative_region_area=try_float_or_nan(relative_region_area),
            region_name=region_name,
            comment=get_val_from_xml_or_none(region_element, "Comment"),
            calculated_data=_get_calculated_data(
                element=region_element,
                excluded_tags=NON_CALCULATED_DATA_TAGS_REGION,
                source_id=region_identifier,
                feature="data region",
            ),
        )


@dataclass(frozen=True)
class Sample:
    measurement_identifier: str
    measurement_time: str
    compartment_temperature: float | None
    location_identifier: str
    sample_identifier: str
    description: str | None
    peak_list: list[Peak]
    data_regions: list[DataRegion]
    calculated_data: list[CalculatedDocument]
    error: str | None

    @staticmethod
    def create(sample_element: ET.Element, screen_tape: ET.Element) -> Sample:
        measurement_id = random_uuid_str()
        well_number = get_val_from_xml(sample_element, "WellNumber")
        screen_tape_id = get_val_from_xml(sample_element, "ScreenTapeID")
        comment = get_val_from_xml_or_none(sample_element, "Comment")
        observations = get_val_from_xml_or_none(sample_element, "Observations")
        description = f"{comment or ''} {observations or ''}".strip()

        peaks = get_element_from_xml(sample_element, "Peaks")
        regions: list[ET.Element] = []

        regions_element = sample_element.find("Regions")
        if regions_element is not None:
            regions = sorted(
                regions_element.iter("Region"),
                key=lambda region: get_val_from_xml(region, "From"),
            )
        error = get_val_from_xml_or_none(sample_element, "Alert")
        if error is not None:
            error = error.strip()

        return Sample(
            measurement_identifier=measurement_id,
            measurement_time=get_val_from_xml(screen_tape, "TapeRunDate"),
            compartment_temperature=try_float_or_none(
                get_val_from_xml_or_none(screen_tape, "ElectrophoresisTemp")
            ),
            location_identifier=well_number,
            sample_identifier=f"{screen_tape_id}_{well_number}",
            description=description or None,
            peak_list=[Peak.create(peak) for peak in peaks.iter("Peak")],
            data_regions=[
                DataRegion.create(region, str(idx))
                for idx, region in enumerate(regions, start=1)
            ],
            calculated_data=_get_calculated_data(
                element=sample_element,
                excluded_tags=NON_CALCULATED_DATA_TAGS_SAMPLE,
                source_id=measurement_id,
                feature="sample",
            ),
            error=error,
        )


@dataclass(frozen=True)
class SamplesList:
    samples: list[Sample]

    @staticmethod
    def create(root_element: ET.Element) -> SamplesList:
        screen_tapes_element = get_element_from_xml(root_element, "ScreenTapes")
        screen_tapes = {
            get_val_from_xml(screen_tape, "ScreenTapeID"): screen_tape
            for screen_tape in screen_tapes_element.iter("ScreenTape")
        }

        samples_element = get_element_from_xml(root_element, "Samples")
        samples = []
        for sample_element in samples_element.iter("Sample"):
            screen_tape_id = get_val_from_xml(sample_element, "ScreenTapeID")
            if screen_tape_id not in screen_tapes:
                msg = NO_SCREEN_TAPE_ID_MATCH.format(screen_tape_id)
                raise AllotropeConversionError(msg)
            samples.append(Sample.create(sample_element, screen_tapes[screen_tape_id]))

        return SamplesList(samples=samples)


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    samples_list: SamplesList

    @staticmethod
    def create(contents: IOType) -> Data:
        try:
            root_element = ET.parse(contents).getroot()  # noqa: S314
        except ET.ParseError as e:
            msg = f"There was an error when trying to read the xml file: {e}"
            raise AllotropeConversionError(message=msg) from e

        return Data(
            metadata=Metadata.create(root_element),
            samples_list=SamplesList.create(root_element),
        )



from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueKiloDalton,
    TQuantityValueNumber,
    TQuantityValueSecondTime,
)

BRAND_NAME = "TapeStation"
DETECTION_TYPE = "fluorescence"
DEVICE_TYPE = "electrophoresis device"
PRODUCT_MANUFACTURER = "Agilent"
SOFTWARE_NAME = "TapeStation Analysis Software"

NO_SCREEN_TAPE_ID_MATCH = "The ScreenTape ID associated with the sample {} does not match any ScreenTape element."

UNIT_CLASSES = type[
    TQuantityValueKiloDalton | TQuantityValueNumber | TQuantityValueSecondTime
]
UNIT_CLASS_LOOKUP: dict[str, UNIT_CLASSES] = {
    "nt": TQuantityValueNumber,
    "bp": TQuantityValueNumber,
    "kD": TQuantityValueKiloDalton,
}
NON_CALCULATED_DATA_TAGS_SAMPLE = [
    "Comment",
    "Observations",
    "Peaks",
    "Regions",
    "ScreenTapeID",
    "WellNumber",
]
NON_CALCULATED_DATA_TAGS_PEAK = [
    "Area",
    "Comment",
    "FromMW",
    "Height",
    "Number",
    "Observations",
    "PercentIntegratedArea",
    "PercentOfTotal",
    "Size",
    "ToMW",
]
NON_CALCULATED_DATA_TAGS_REGION = [
    "From",
    "To",
    "Area",
    "PercentOfTotal",
    "Comment",
]



# header
# sample setup
# raw data
# amplification data
# multicomponent data
# results
# melt curve raw data

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from io import StringIO
import re

import numpy as np
import pandas as pd

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ExperimentType
from allotropy.allotrope.pandas_util import read_csv
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    Referenceable,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_empty_df,
    assert_not_none,
    df_to_series,
    try_bool_from_series_or_none,
    try_float_from_series,
    try_float_from_series_or_none,
    try_float_or_none,
    try_int,
    try_int_from_series,
    try_str_from_series,
    try_str_from_series_or_default,
    try_str_from_series_or_none,
)


@dataclass(frozen=True)
class Header:
    measurement_time: str
    plate_well_count: int | None
    experiment_type: ExperimentType
    device_identifier: str
    model_number: str
    device_serial_number: str
    measurement_method_identifier: str
    pcr_detection_chemistry: str
    passive_reference_dye_setting: str | None
    barcode: str | None
    analyst: str | None
    experimental_data_identifier: str | None

    @staticmethod
    def create(reader: LinesReader) -> Header:
        lines = [line.replace("*", "", 1) for line in reader.pop_until(r"^\[.+\]")]
        csv_stream = StringIO("\n".join(lines))
        raw_data = read_csv(csv_stream, header=None, sep="=", names=["index", "values"])
        data = pd.Series(raw_data["values"].values, index=raw_data["index"])
        data.index = data.index.str.strip()
        data = data.str.strip().replace("NA", None)

        experiments_type_options = {
            "Standard Curve": ExperimentType.standard_curve_qPCR_experiment,
            "Relative Standard Curve": ExperimentType.relative_standard_curve_qPCR_experiment,
            "Comparative Cт (ΔΔCт)": ExperimentType.comparative_CT_qPCR_experiment,
            "Melt Curve": ExperimentType.melt_curve_qPCR_experiment,
            "Genotyping": ExperimentType.genotyping_qPCR_experiment,
            "Presence/Absence": ExperimentType.presence_absence_qPCR_experiment,
        }

        plate_well_count_search = re.search(
            "(96)|(384)",
            try_str_from_series(data, "Block Type"),
        )

        return Header(
            measurement_time=try_str_from_series(data, "Experiment Run End Time"),
            plate_well_count=(
                None
                if plate_well_count_search is None
                else int(plate_well_count_search.group())
            ),
            experiment_type=assert_not_none(
                experiments_type_options.get(
                    try_str_from_series(data, "Experiment Type"),
                ),
                msg="Unable to find valid experiment type",
            ),
            device_identifier=(
                try_str_from_series_or_none(data, "Instrument Name") or NOT_APPLICABLE
            ),
            model_number=try_str_from_series(data, "Instrument Type"),
            device_serial_number=try_str_from_series_or_none(
                data, "Instrument Serial Number"
            )
            or NOT_APPLICABLE,
            measurement_method_identifier=try_str_from_series(
                data, "Quantification Cycle Method"
            ),
            pcr_detection_chemistry=try_str_from_series(data, "Chemistry"),
            passive_reference_dye_setting=try_str_from_series_or_none(
                data, "Passive Reference"
            ),
            barcode=try_str_from_series_or_none(data, "Experiment Barcode"),
            analyst=try_str_from_series_or_none(data, "Experiment User Name"),
            experimental_data_identifier=try_str_from_series_or_none(
                data, "Experiment Name"
            ),
        )


@dataclass
class WellItem(Referenceable):
    identifier: int
    target_dna_description: str
    sample_identifier: str
    reporter_dye_setting: str | None
    position: str | None
    well_location_identifier: str | None
    quencher_dye_setting: str | None
    sample_role_type: str | None
    _amplification_data: AmplificationData | None = None
    _result: Result | None = None

    @property
    def amplification_data(self) -> AmplificationData:
        return assert_not_none(
            self._amplification_data,
            msg=f"Unable to find amplification data for target '{self.target_dna_description}' in well {self.identifier} .",
        )

    @amplification_data.setter
    def amplification_data(self, amplification_data: AmplificationData) -> None:
        self._amplification_data = amplification_data

    @property
    def result(self) -> Result:
        return assert_not_none(
            self._result,
            msg=f"Unable to find result data for well {self.identifier}.",
        )

    @result.setter
    def result(self, result: Result) -> None:
        self._result = result

    @staticmethod
    def create_genotyping(data: pd.Series[str]) -> tuple[WellItem, WellItem]:
        identifier = try_int_from_series(data, "Well")

        snp_name = try_str_from_series(
            data,
            "SNP Assay Name",
            msg=f"Unable to find snp name for well {identifier}",
        )

        sample_identifier = try_str_from_series_or_default(
            data,
            "Sample Name",
            default=NOT_APPLICABLE,
        )

        allele1 = try_str_from_series(
            data,
            "Allele1 Name",
            msg=f"Unable to find allele 1 for well {identifier}",
        )

        allele2 = try_str_from_series(
            data,
            "Allele2 Name",
            msg=f"Unable to find allele 2 for well {identifier}",
        )

        return (
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele1}",
                sample_identifier=sample_identifier,
                reporter_dye_setting=try_str_from_series_or_none(
                    data, "Allele1 Reporter"
                ),
                position=try_str_from_series_or_default(
                    data, "Well Position", default=NOT_APPLICABLE
                ),
                well_location_identifier=try_str_from_series_or_none(
                    data, "Well Position"
                ),
                quencher_dye_setting=try_str_from_series_or_none(data, "Quencher"),
                sample_role_type=try_str_from_series_or_none(data, "Task"),
            ),
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele2}",
                sample_identifier=sample_identifier,
                reporter_dye_setting=try_str_from_series_or_none(
                    data, "Allele2 Reporter"
                ),
                position=try_str_from_series_or_default(
                    data, "Well Position", default=NOT_APPLICABLE
                ),
                well_location_identifier=try_str_from_series_or_none(
                    data, "Well Position"
                ),
                quencher_dye_setting=try_str_from_series_or_none(data, "Quencher"),
                sample_role_type=try_str_from_series_or_none(data, "Task"),
            ),
        )

    @staticmethod
    def create_generic(data: pd.Series[str]) -> WellItem:
        identifier = try_int_from_series(data, "Well")

        target_dna_description = try_str_from_series(
            data,
            "Target Name",
            msg=f"Unable to find target dna description for well {identifier}",
        )

        sample_identifier = try_str_from_series_or_default(
            data,
            "Sample Name",
            default=NOT_APPLICABLE,
        )

        return WellItem(
            uuid=random_uuid_str(),
            identifier=identifier,
            target_dna_description=target_dna_description,
            sample_identifier=sample_identifier,
            reporter_dye_setting=try_str_from_series_or_none(data, "Reporter"),
            position=try_str_from_series_or_default(
                data, "Well Position", default=NOT_APPLICABLE
            ),
            well_location_identifier=try_str_from_series_or_none(data, "Well Position"),
            quencher_dye_setting=try_str_from_series_or_none(data, "Quencher"),
            sample_role_type=try_str_from_series_or_none(data, "Task"),
        )


@dataclass
class Well:
    identifier: int
    items: dict[str, WellItem]
    _multicomponent_data: MulticomponentData | None = None
    _melt_curve_raw_data: MeltCurveRawData | None = None

    def get_well_item(self, target: str) -> WellItem:
        well_item = self.items.get(target)
        return assert_not_none(
            well_item,
            msg=f"Unable to find target DNA '{target}' for well {self.identifier}.",
        )

    @property
    def multicomponent_data(self) -> MulticomponentData | None:
        return self._multicomponent_data

    @multicomponent_data.setter
    def multicomponent_data(self, multicomponent_data: MulticomponentData) -> None:
        self._multicomponent_data = multicomponent_data

    @property
    def melt_curve_raw_data(self) -> MeltCurveRawData | None:
        return self._melt_curve_raw_data

    @melt_curve_raw_data.setter
    def melt_curve_raw_data(self, melt_curve_raw_data: MeltCurveRawData) -> None:
        self._melt_curve_raw_data = melt_curve_raw_data

    @staticmethod
    def create_genotyping(identifier: int, well_data: pd.Series[str]) -> Well:
        return Well(
            identifier=identifier,
            items={
                well_item.target_dna_description: well_item
                for well_item in WellItem.create_genotyping(well_data)
            },
        )

    @staticmethod
    def create_generic(identifier: int, well_data: pd.DataFrame) -> Well:
        return Well(
            identifier=identifier,
            items={
                item_data["Target Name"]: WellItem.create_generic(item_data)
                for _, item_data in well_data.iterrows()
            },
        )


@dataclass(frozen=True)
class WellList:
    wells: list[Well]

    def iter_well_items(self) -> Iterator[WellItem]:
        for well in self.wells:
            yield from well.items.values()

    def __iter__(self) -> Iterator[Well]:
        return iter(self.wells)

    @staticmethod
    def create(reader: LinesReader, experiment_type: ExperimentType) -> WellList:
        assert_not_none(
            reader.drop_until(r"^\[Sample Setup\]"),
            msg="Unable to find 'Sample Setup' section in file.",
        )

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        data = read_csv(csv_stream, sep="\t").replace(np.nan, None)

        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            return WellList(
                [
                    Well.create_genotyping(
                        try_int(str(identifier), "genotyping well identifier"),
                        well_data,
                    )
                    for identifier, well_data in data.iterrows()
                ]
            )
        return WellList(
            [
                Well.create_generic(
                    try_int(str(identifier), "well identifier"),
                    well_data,
                )
                for identifier, well_data in data[
                    data["Target Name"].notnull()
                ].groupby("Well")
            ]
        )


@dataclass(frozen=True)
class RawData:
    lines: list[str]

    @staticmethod
    def create(reader: LinesReader) -> RawData | None:
        if reader.match(r"^\[Raw Data\]"):
            reader.pop()  # remove title
            return RawData(lines=list(reader.pop_until(r"^\[.+\]")))
        return None


@dataclass(frozen=True)
class AmplificationData:
    total_cycle_number_setting: float
    cycle: list[float]
    rn: list[float | None]
    delta_rn: list[float | None]

    @staticmethod
    def get_data(reader: LinesReader) -> pd.DataFrame:
        assert_not_none(
            reader.drop_until(r"^\[Amplification Data\]"),
            msg="Unable to find 'Amplification Data' section in file.",
        )

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        return read_csv(csv_stream, sep="\t", thousands=r",")

    @staticmethod
    def create(
        amplification_data: pd.DataFrame, well_item: WellItem
    ) -> AmplificationData:
        well_data = assert_not_empty_df(
            amplification_data[amplification_data["Well"] == well_item.identifier],
            msg=f"Unable to find amplification data for well {well_item.identifier}.",
        )

        target_data = assert_not_empty_df(
            well_data[well_data["Target Name"] == well_item.target_dna_description],
            msg=f"Unable to find amplification data for target '{well_item.target_dna_description}' in well {well_item.identifier} .",
        )

        return AmplificationData(
            total_cycle_number_setting=float(target_data["Cycle"].max()),
            cycle=target_data["Cycle"].tolist(),
            rn=target_data["Rn"].tolist(),
            delta_rn=target_data["Delta Rn"].tolist(),
        )


@dataclass(frozen=True)
class MulticomponentData:
    cycle: list[float]
    columns: dict[str, list[float | None]]

    def get_column(self, name: str) -> list[float | None]:
        return assert_not_none(
            self.columns.get(name),
            msg=f"Unable to obtain '{name}' from multicomponent data.",
        )

    @staticmethod
    def get_data(reader: LinesReader) -> pd.DataFrame | None:
        if not reader.match(r"^\[Multicomponent Data\]"):
            return None
        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        return read_csv(csv_stream, sep="\t", thousands=r",")

    @staticmethod
    def create(data: pd.DataFrame, well: Well) -> MulticomponentData:
        well_data = assert_not_empty_df(
            data[data["Well"] == well.identifier],
            msg=f"Unable to find multi component data for well {well.identifier}.",
        )

        return MulticomponentData(
            cycle=well_data["Cycle"].tolist(),
            columns={
                name: well_data[name].tolist()  # type: ignore[misc]
                for name in well_data
                if name not in ["Well", "Cycle", "Well Position"]
            },
        )


@dataclass(frozen=True)
class Result:
    cycle_threshold_value_setting: float
    cycle_threshold_result: float | None
    automatic_cycle_threshold_enabled_setting: bool | None
    automatic_baseline_determination_enabled_setting: bool | None
    normalized_reporter_result: float | None
    baseline_corrected_reporter_result: float | None
    genotyping_determination_result: str | None
    genotyping_determination_method_setting: float | None
    quantity: float | None
    quantity_mean: float | None
    quantity_sd: float | None
    ct_mean: float | None
    ct_sd: float | None
    delta_ct_mean: float | None
    delta_ct_se: float | None
    delta_delta_ct: float | None
    rq: float | None
    rq_min: float | None
    rq_max: float | None
    rn_mean: float | None
    rn_sd: float | None
    y_intercept: float | None
    r_squared: float | None
    slope: float | None
    efficiency: float | None

    @staticmethod
    def get_data(reader: LinesReader) -> tuple[pd.DataFrame, pd.Series[str]]:
        assert_not_none(
            reader.drop_until(r"^\[Results\]"),
            msg="Unable to find 'Results' section in file.",
        )

        reader.pop()  # remove title
        data_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(data_lines))
        data = read_csv(csv_stream, sep="\t", thousands=r",").replace(np.nan, None)

        reader.drop_empty()

        if reader.match(r"\[.+\]"):
            return data, pd.Series()

        metadata_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(metadata_lines))
        raw_data = read_csv(csv_stream, header=None, sep="=", names=["index", "values"])
        metadata = pd.Series(raw_data["values"].values, index=raw_data["index"])
        metadata.index = metadata.index.str.strip()

        reader.drop_empty()

        return data, metadata.str.strip()

    @staticmethod
    def create_genotyping(data: pd.DataFrame, well_item: WellItem) -> Result:
        well_data = assert_not_empty_df(
            data[data["Well"] == well_item.identifier],
            msg=f"Unable to find result data for well {well_item.identifier}.",
        )

        snp_assay_name, _ = well_item.target_dna_description.split("-")
        target_data = df_to_series(
            assert_not_empty_df(
                well_data[well_data["SNP Assay Name"] == snp_assay_name],
                msg=f"Unable to find result data for well {well_item.identifier}.",
            ),
            msg=f"Expected exactly 1 row of results to be associated with target '{well_item.target_dna_description}' in well {well_item.identifier}.",
        )

        _, raw_allele = well_item.target_dna_description.split("-")
        allele = raw_allele.replace(" ", "")
        cycle_threshold_value_setting = try_float_from_series(
            target_data,
            f"{allele} Ct Threshold",
            msg=f"Unable to find cycle threshold value setting for well {well_item.identifier}",
        )

        cycle_threshold_result = assert_not_none(
            target_data.get(f"{allele} Ct"),
            msg="Unable to find cycle threshold result",
        )

        return Result(
            cycle_threshold_value_setting=cycle_threshold_value_setting,
            cycle_threshold_result=try_float_or_none(str(cycle_threshold_result)),
            automatic_cycle_threshold_enabled_setting=try_bool_from_series_or_none(
                target_data, f"{allele} Automatic Ct Threshold"
            ),
            automatic_baseline_determination_enabled_setting=try_bool_from_series_or_none(
                target_data, f"{allele} Automatic Baseline"
            ),
            normalized_reporter_result=try_float_from_series_or_none(target_data, "Rn"),
            baseline_corrected_reporter_result=try_float_from_series_or_none(
                target_data, f"{allele} Delta Rn"
            ),
            genotyping_determination_result=try_str_from_series_or_none(
                target_data, "Call"
            ),
            genotyping_determination_method_setting=try_float_from_series_or_none(
                target_data, "Threshold Value"
            ),
            quantity=try_float_from_series_or_none(target_data, "Quantity"),
            quantity_mean=try_float_from_series_or_none(target_data, "Quantity Mean"),
            quantity_sd=try_float_from_series_or_none(target_data, "Quantity SD"),
            ct_mean=try_float_from_series_or_none(target_data, "Ct Mean"),
            ct_sd=try_float_from_series_or_none(target_data, "Ct SD"),
            delta_ct_mean=try_float_from_series_or_none(target_data, "Delta Ct Mean"),
            delta_ct_se=try_float_from_series_or_none(target_data, "Delta Ct SE"),
            delta_delta_ct=try_float_from_series_or_none(target_data, "Delta Delta Ct"),
            rq=try_float_from_series_or_none(target_data, "RQ"),
            rq_min=try_float_from_series_or_none(target_data, "RQ Min"),
            rq_max=try_float_from_series_or_none(target_data, "RQ Max"),
            rn_mean=try_float_from_series_or_none(target_data, "Rn Mean"),
            rn_sd=try_float_from_series_or_none(target_data, "Rn SD"),
            y_intercept=try_float_from_series_or_none(target_data, "Y-Intercept"),
            r_squared=try_float_from_series_or_none(target_data, "R(superscript 2)"),
            slope=try_float_from_series_or_none(target_data, "Slope"),
            efficiency=try_float_from_series_or_none(target_data, "Efficiency"),
        )

    @staticmethod
    def create_generic(data: pd.DataFrame, well_item: WellItem) -> Result:
        well_data = assert_not_empty_df(
            data[data["Well"] == well_item.identifier],
            msg=f"Unable to find result data for well {well_item.identifier}.",
        )

        target_data = df_to_series(
            assert_not_empty_df(
                well_data[well_data["Target Name"] == well_item.target_dna_description],
                msg=f"Unable to find result data for well {well_item.identifier}.",
            ),
            msg=f"Expected exactly 1 row of results to be associated with target '{well_item.target_dna_description}' in well {well_item.identifier}.",
        )

        cycle_threshold_result = assert_not_none(
            target_data.get("CT"),
            msg="Unable to find cycle threshold result",
        )

        return Result(
            cycle_threshold_value_setting=try_float_from_series(
                target_data,
                "Ct Threshold",
                msg=f"Unable to find cycle threshold value setting for well {well_item.identifier}",
            ),
            cycle_threshold_result=try_float_or_none(str(cycle_threshold_result)),
            automatic_cycle_threshold_enabled_setting=try_bool_from_series_or_none(
                target_data, "Automatic Ct Threshold"
            ),
            automatic_baseline_determination_enabled_setting=try_bool_from_series_or_none(
                target_data, "Automatic Baseline"
            ),
            normalized_reporter_result=try_float_from_series_or_none(target_data, "Rn"),
            baseline_corrected_reporter_result=try_float_from_series_or_none(
                target_data, "Delta Rn"
            ),
            genotyping_determination_result=try_str_from_series_or_none(
                target_data, "Call"
            ),
            genotyping_determination_method_setting=try_float_from_series_or_none(
                target_data, "Threshold Value"
            ),
            quantity=try_float_from_series_or_none(target_data, "Quantity"),
            quantity_mean=try_float_from_series_or_none(target_data, "Quantity Mean"),
            quantity_sd=try_float_from_series_or_none(target_data, "Quantity SD"),
            ct_mean=try_float_from_series_or_none(target_data, "Ct Mean"),
            ct_sd=try_float_from_series_or_none(target_data, "Ct SD"),
            delta_ct_mean=try_float_from_series_or_none(target_data, "Delta Ct Mean"),
            delta_ct_se=try_float_from_series_or_none(target_data, "Delta Ct SE"),
            delta_delta_ct=try_float_from_series_or_none(target_data, "Delta Delta Ct"),
            rq=try_float_from_series_or_none(target_data, "RQ"),
            rq_min=try_float_from_series_or_none(target_data, "RQ Min"),
            rq_max=try_float_from_series_or_none(target_data, "RQ Max"),
            rn_mean=try_float_from_series_or_none(target_data, "Rn Mean"),
            rn_sd=try_float_from_series_or_none(target_data, "Rn SD"),
            y_intercept=try_float_from_series_or_none(target_data, "Y-Intercept"),
            r_squared=try_float_from_series_or_none(target_data, "R(superscript 2)"),
            slope=try_float_from_series_or_none(target_data, "Slope"),
            efficiency=try_float_from_series_or_none(target_data, "Efficiency"),
        )

    @staticmethod
    def create(
        data: pd.DataFrame, well_item: WellItem, experiment_type: ExperimentType
    ) -> Result:
        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            return Result.create_genotyping(data, well_item)
        return Result.create_generic(data, well_item)


@dataclass(frozen=True)
class MeltCurveRawData:
    reading: list[float]
    fluorescence: list[float | None]
    derivative: list[float | None]

    @staticmethod
    def create(data: pd.DataFrame, well: Well) -> MeltCurveRawData:
        well_data = assert_not_empty_df(
            data[data["Well"] == well.identifier],
            msg=f"Unable to find melt curve raw data for well {well.identifier}.",
        )
        return MeltCurveRawData(
            reading=well_data["Reading"].tolist(),
            fluorescence=well_data["Fluorescence"].tolist(),
            derivative=well_data["Derivative"].tolist(),
        )

    @staticmethod
    def get_data(reader: LinesReader) -> pd.DataFrame | None:
        if not reader.match(r"^\[Melt Curve Raw Data\]"):
            return None
        reader.pop()  # remove title
        lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(lines))
        return read_csv(csv_stream, sep="\t", thousands=r",")


@dataclass(frozen=True)
class Data:
    header: Header
    wells: WellList
    raw_data: RawData | None
    endogenous_control: str
    reference_sample: str
    calculated_documents: list[CalculatedDocument]
